<script type="module">
    import { Lona } from '{{ Lona.settings.STATIC_URL_PREFIX }}/_lona/lona.js';

    Lona.set_protocol(JSON.parse('{{ Lona.get_protocol_as_json() }}'));
    Lona.set_settings(JSON.parse('{{ Lona.get_settings_as_json() }}'));

    window.Lona = Lona;

</script>

{% for script in scripts %}<script src="{{ Lona.settings.STATIC_URL_PREFIX }}{{ script.url }}" defer></script>
{% endfor -%}
